#pragma once

#include <algorithm>
#include <cassert>

// Master hash function: Bob Jenkins' SpookyHash
#include "SpookyHash/SpookyV2.h"
inline size_t HashMemory(void * p, size_t sizeBytes)
{
	return size_t(SpookyHash::Hash64(p, sizeBytes, 0));
}

static const int s_hashTableInitialSize = 16;



// C0HashTable implementation

template <typename K, typename V>
C0HashTable<K, V>::C0HashTable()
:	size(0)
{
	// Start off with a small initial size
	buckets.resize(s_hashTableInitialSize);
	elemPool.resize(s_hashTableInitialSize);

	// Build the initial free list
	pElemFreeHead = &elemPool[0];
	for (int i = 0; i < s_hashTableInitialSize - 1; ++i)
		elemPool[i].pNext = &elemPool[i+1];
}

template <typename K, typename V>
void C0HashTable<K, V>::Insert(K key, V value)
{
	// Resize larger if we're out of elements
	if (!pElemFreeHead)
	{
		Rehash(buckets.size() * 2);
	}

	// Grab the next element off the free list
	assert(pElemFreeHead);
	Elem * e = pElemFreeHead;
	pElemFreeHead = e->pNext;

	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key);
	Bucket * b = &buckets[hash % buckets.size()];

	// Insert the new element into the bucket
	e->pNext = b->pHead;
	b->pHead = e;

	// Store the hash, key, and value in the element
	e->hash = hash;
	e->key = key;
	e->value = value;

	++size;
}

template <typename K, typename V>
V * C0HashTable<K, V>::Lookup(K key)
{
	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key);
	Bucket * b = &buckets[hash % buckets.size()];

	// Walk the chain looking for a matching element
	for (Elem * e = b->pHead; e; e = e->pNext)
	{
		if (e->hash == hash && e->key == key)
			return &e->value;
	}

	return nullptr;
}

template <typename K, typename V>
bool C0HashTable<K, V>::Remove(K key)
{
	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key);
	Bucket * b = &buckets[hash % buckets.size()];

	Elem * pHead = b->pHead;
	if (!pHead)
		return false;

	Elem * eRemoved = nullptr;

	// Check the first element
	if (pHead->hash == hash && pHead->key == key)
	{
		eRemoved = pHead;
		b->pHead = pHead->pNext;
	}
	else
	{
		// Walk the chain looking for a matching element
		for (Elem * e = pHead->pNext, * ePrev = pHead;
			 e;
			 ePrev = e, e = e->pNext)
		{
			if (e->hash == hash && e->key == key)
			{
				eRemoved = e;
				ePrev->pNext = e->pNext;
				break;
			}
		}
	}

	if (eRemoved)
	{
		// Put eRemoved back on the free list
		eRemoved->hash = 0;
		eRemoved->pNext = pElemFreeHead;
		pElemFreeHead = eRemoved;
		--size;
	}

	return (eRemoved != nullptr);
}

template <typename K, typename V>
void C0HashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize);
}

template <typename K, typename V>
void C0HashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets and elements
	std::vector<Bucket> bucketsNew(bucketCountNew);
	std::vector<Elem> elemPoolNew(bucketCountNew);

	// Walk through all the current elements, move them into the new
	// element pool and insert them into the new buckets
	size_t iElemNext = 0;
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		for (Elem * e = b->pHead; e; e = e->pNext)
		{
			Elem * eNew = &elemPoolNew[iElemNext];
			++iElemNext;

			size_t hash = e->hash;
			Bucket * bNew = &bucketsNew[hash % bucketCountNew];

			// Insert the new element into the bucket
			eNew->pNext = bNew->pHead;
			bNew->pHead = eNew;

			// Store the hash, key, and value in the element
			eNew->hash = hash;
			eNew->key = std::move(e->key);
			eNew->value = std::move(e->value);
		}
	}

	// Swap the new buckets and elements into place
	buckets.swap(bucketsNew);
	elemPool.swap(elemPoolNew);

	// Build a free list of the remaining elements in the pool
	pElemFreeHead = (iElemNext < elemPool.size()) ? &elemPool[iElemNext] : nullptr;
	for (size_t i = iElemNext, iEnd = elemPool.size() - 1; i < iEnd; ++i)
		elemPool[i].pNext = &elemPool[i+1];
}

template <typename K, typename V>
void C0HashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);
	elemPool.clear();
	elemPool.resize(s_hashTableInitialSize);

	// Build the initial free list
	pElemFreeHead = &elemPool[0];
	for (int i = 0; i < s_hashTableInitialSize - 1; ++i)
		elemPool[i].pNext = &elemPool[i+1];

	size = 0;
}



// C1HashTable implementation

static const size_t s_63Bits = 0x7fffffffffffffffULL;

template <typename K, typename V>
C1HashTable<K, V>::C1HashTable()
:	size(0)
{
	// Start off with a small initial size.  Since we have space for an
	// element in the bucket itself, start with only half as many elements
	// in the element pool.
	buckets.resize(s_hashTableInitialSize);
	elemPool.resize(s_hashTableInitialSize / 2);

	// Build the initial free list
	pElemFreeHead = &elemPool[0];
	for (size_t i = 0, iEnd = elemPool.size() - 1; i < iEnd; ++i)
		elemPool[i].pNext = &elemPool[i+1];
}

template <typename K, typename V>
void C1HashTable<K, V>::Insert(K key, V value)
{
	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key) & s_63Bits;
	Bucket * b = &buckets[hash % buckets.size()];

	// Is the bucket empty?
	if (!b->filled)
	{
		// Store it in the bucket. Done.
		b->filled = true;
		b->hash = hash;
		b->key = key;
		b->value = value;
		++size;
		return;
	}

	// Resize larger if we're out of elements
	if (!pElemFreeHead)
	{
		Rehash(buckets.size() * 2);

		// Need to re-lookup the bucket since we resized
		b = &buckets[hash % buckets.size()];

		// Is the bucket empty?
		if (!b->filled)
		{
			// Store it in the bucket. Done.
			b->filled = true;
			b->hash = hash;
			b->key = key;
			b->value = value;
			++size;
			return;
		}
	}

	// Grab the next element off the free list
	assert(pElemFreeHead);
	Elem * e = pElemFreeHead;
	pElemFreeHead = e->pNext;

	// Insert the new element into the bucket
	e->pNext = b->pHead;
	b->pHead = e;

	// Store the hash, key, and value in the element
	e->hash = hash;
	e->key = key;
	e->value = value;

	++size;
}

template <typename K, typename V>
V * C1HashTable<K, V>::Lookup(K key)
{
	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key) & s_63Bits;
	Bucket * b = &buckets[hash % buckets.size()];

	if (!b->filled)
		return nullptr;

	// Check if it's in the bucket itself
	if (b->hash == hash && b->key == key)
	{
		return &b->value;
	}

	// Walk the chain looking for a matching element
	for (Elem * e = b->pHead; e; e = e->pNext)
	{
		if (e->hash == hash && e->key == key)
			return &e->value;
	}

	return nullptr;
}

template <typename K, typename V>
bool C1HashTable<K, V>::Remove(K key)
{
	// Hash the key and look up the appropriate bucket
	size_t hash = HashKey(key) & s_63Bits;
	Bucket * b = &buckets[hash % buckets.size()];

	if (!b->filled)
		return nullptr;

	// Check if it's in the bucket itself
	if (b->hash == hash && b->key == key)
	{
		// If the bucket has a chain, move the first element of the chain
		// into the bucket
		if (Elem * pHead = b->pHead)
		{
			b->pHead = pHead->pNext;
			b->hash = pHead->hash;
			b->key = std::move(pHead->key);
			b->value = std::move(pHead->value);

			// Put the removed element back on the free list
			pHead->hash = 0;
			pHead->pNext = pElemFreeHead;
			pElemFreeHead = pHead;
		}
		else
		{
			// The bucket is now empty
			b->hash = 0;
			b->filled = false;
		}

		--size;
		return true;
	}

	Elem * pHead = b->pHead;
	if (!pHead)
		return false;

	Elem * eRemoved = nullptr;

	// Check the first element
	if (pHead->hash == hash && pHead->key == key)
	{
		eRemoved = pHead;
		b->pHead = pHead->pNext;
	}
	else
	{
		// Walk the chain looking for a matching element
		for (Elem * e = pHead->pNext, * ePrev = pHead;
			 e;
			 ePrev = e, e = e->pNext)
		{
			if (e->hash == hash && e->key == key)
			{
				eRemoved = e;
				ePrev->pNext = e->pNext;
				break;
			}
		}
	}

	if (eRemoved)
	{
		// Put eRemoved back on the free list
		eRemoved->hash = 0;
		eRemoved->pNext = pElemFreeHead;
		pElemFreeHead = eRemoved;
		--size;
	}

	return (eRemoved != nullptr);
}

template <typename K, typename V>
void C1HashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize);
}

template <typename K, typename V>
void C1HashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets and elements
	std::vector<Bucket> bucketsNew(bucketCountNew);
	std::vector<Elem> elemPoolNew(bucketCountNew / 2);

	// Walk through all the current elements, move them into the new
	// element pool and insert them into the new buckets
	size_t iElemNext = 0;
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];

		if (!b->filled)
			continue;

		// Handle the element in the bucket
		size_t hash = b->hash;
		Bucket * bNew = &bucketsNew[hash % bucketCountNew];
		if (!bNew->filled)
		{
			// Store it in the bucket. Done.
			bNew->filled = true;
			bNew->hash = hash;
			// Note: can't move from old key and value because we might
			// have to use the escape hatch (rehash even bigger) later.
			bNew->key = b->key;
			bNew->value = b->value;
		}
		else
		{
			if (iElemNext == elemPoolNew.size())
			{
				// Escape hatch: rehash even bigger!
				bucketsNew.clear();
				elemPoolNew.clear();
				Rehash(bucketCountNew * 2);
				return;
			}
			Elem * eNew = &elemPoolNew[iElemNext];
			++iElemNext;

			// Insert the new element into the bucket
			eNew->pNext = bNew->pHead;
			bNew->pHead = eNew;

			// Store the hash, key, and value in the element
			eNew->hash = hash;
			// Note: can't move from old key and value because we might
			// have to use the escape hatch (rehash even bigger) later.
			eNew->key = b->key;
			eNew->value = b->value;
		}

		// Handle the elements in the chain
		for (Elem * e = b->pHead; e; e = e->pNext)
		{
			size_t hash = e->hash;
			Bucket * bNew = &bucketsNew[hash % bucketCountNew];

			if (!bNew->filled)
			{
				// Store it in the bucket. Done.
				bNew->filled = true;
				bNew->hash = hash;
				// Note: can't move from old key and value because we might
				// have to use the escape hatch (rehash even bigger) later.
				bNew->key = e->key;
				bNew->value = e->value;
			}
			else
			{
				if (iElemNext == elemPoolNew.size())
				{
					// Escape hatch: rehash even bigger!
					bucketsNew.clear();
					elemPoolNew.clear();
					Rehash(bucketCountNew * 2);
					return;
				}
				Elem * eNew = &elemPoolNew[iElemNext];
				++iElemNext;

				// Insert the new element into the bucket
				eNew->pNext = bNew->pHead;
				bNew->pHead = eNew;

				// Store the hash, key, and value in the element
				eNew->hash = hash;
				// Note: can't move from old key and value because we might
				// have to use the escape hatch (rehash even bigger) later.
				eNew->key = e->key;
				eNew->value = e->value;
			}
		}
	}

	// Swap the new buckets and elements into place
	buckets.swap(bucketsNew);
	elemPool.swap(elemPoolNew);

	// Build a free list of the remaining elements in the pool
	pElemFreeHead = (iElemNext < elemPool.size()) ? &elemPool[iElemNext] : nullptr;
	for (size_t i = iElemNext, iEnd = elemPool.size() - 1; i < iEnd; ++i)
		elemPool[i].pNext = &elemPool[i+1];
}

template <typename K, typename V>
void C1HashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);
	elemPool.clear();
	elemPool.resize(s_hashTableInitialSize / 2);

	// Build the initial free list
	pElemFreeHead = &elemPool[0];
	for (size_t i = 0, iEnd = elemPool.size() - 1; i < iEnd; ++i)
		elemPool[i].pNext = &elemPool[i+1];

	size = 0;
}



// OLHashTable implementation

static const size_t s_62Bits = 0x3fffffffffffffffULL;

template <typename K, typename V>
OLHashTable<K, V>::OLHashTable()
:	size(0)
{
	// Start off with a small initial size
	buckets.resize(s_hashTableInitialSize);
}

template <typename K, typename V>
void OLHashTable<K, V>::Insert(K key, V value)
{
	// Resize larger if the load factor goes over 2/3
	if (size * 3 > buckets.size() * 2)
	{
		Rehash(buckets.size() * 2);
	}

	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search for an unused bucket
	Bucket * bTarget = nullptr;
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i) 
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
		{
			bTarget = b;
			break;
		}
	}
	if (!bTarget)
	{
		for (size_t i = 0; i < iBucketStart; ++i)
		{
			Bucket * b = &buckets[i];
			if (b->state != BSTATE_Filled)
			{
				bTarget = b;
				break;
			}
		}
	}

	assert(bTarget);

	// Store the hash, key, and value in the bucket
	bTarget->hash = hash;
	bTarget->state = BSTATE_Filled;
	bTarget->key = key;
	bTarget->value = value;

	++size;
}

template <typename K, typename V>
V * OLHashTable<K, V>::Lookup(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
				return &b->value;
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
				return &b->value;
			break;
		default:
			break;
		}
	}

	return nullptr;
}

template <typename K, typename V>
bool OLHashTable<K, V>::Remove(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
			{
				b->hash = 0;
				b->state = BSTATE_Removed;
				--size;
				return true;
			}
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
			{
				b->hash = 0;
				b->state = BSTATE_Removed;
				--size;
				return true;
			}
			break;
		default:
			break;
		}
	}

	return false;
}

template <typename K, typename V>
void OLHashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize * 3 / 2);
}

template <typename K, typename V>
void OLHashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets
	std::vector<Bucket> bucketsNew(bucketCountNew);

	// Walk through all the current elements and insert them into the new buckets
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
			continue;

		// Hash the key and find the starting bucket
		size_t hash = b->hash;
		size_t iBucketStart = hash % bucketCountNew;

		// Search for an unused bucket
		Bucket * bTarget = nullptr;
		for (size_t j = iBucketStart; j < bucketCountNew; ++j) 
		{
			Bucket * bNew = &bucketsNew[j];
			if (bNew->state != BSTATE_Filled)
			{
				bTarget = bNew;
				break;
			}
		}
		if (!bTarget)
		{
			for (size_t j = 0; j < iBucketStart; ++j)
			{
				Bucket * bNew = &bucketsNew[j];
				if (bNew->state != BSTATE_Filled)
				{
					bTarget = bNew;
					break;
				}
			}
		}

		assert(bTarget);

		// Store the hash, key, and value in the bucket
		bTarget->hash = hash;
		bTarget->state = BSTATE_Filled;
		bTarget->key = std::move(b->key);
		bTarget->value = std::move(b->value);
	}

	// Swap the new buckets into place
	buckets.swap(bucketsNew);
}

template <typename K, typename V>
void OLHashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);

	size = 0;
}



// OQHashTable implementation

template <typename K, typename V>
OQHashTable<K, V>::OQHashTable()
:	size(0)
{
	// Start off with a small initial size
	buckets.resize(s_hashTableInitialSize);
}

template <typename K, typename V>
void OQHashTable<K, V>::Insert(K key, V value)
{
	// Resize larger if the load factor goes over 2/3
	if (size * 3 > buckets.size() * 2)
	{
		Rehash(buckets.size() * 2);
	}

	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search for an unused bucket
	Bucket * bTarget = nullptr;
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		// Evaluate the probing sequence.  This one is guaranteed to hit all
		// elements if the table is power-of-two-sized; see:
		//   http://en.wikipedia.org/wiki/Quadratic_probing
		size_t probe = (iBucketStart + (i + i*i) / 2) % iEnd;

		Bucket * b = &buckets[probe];
		if (b->state != BSTATE_Filled)
		{
			bTarget = b;
			break;
		}
	}

	assert(bTarget);

	// Store the hash, key, and value in the bucket
	bTarget->hash = hash;
	bTarget->state = BSTATE_Filled;
	bTarget->key = key;
	bTarget->value = value;

	++size;
}

template <typename K, typename V>
V * OQHashTable<K, V>::Lookup(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		// Evaluate the probing sequence
		size_t probe = (iBucketStart + (i + i*i) / 2) % iEnd;

		Bucket * b = &buckets[probe];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
				return &b->value;
			break;
		default:
			break;
		}
	}

	return nullptr;
}

template <typename K, typename V>
bool OQHashTable<K, V>::Remove(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		// Evaluate the probing sequence
		size_t probe = (iBucketStart + (i + i*i) / 2) % iEnd;

		Bucket * b = &buckets[probe];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash && b->key == key)
			{
				b->hash = 0;
				b->state = BSTATE_Removed;
				--size;
				return true;
			}
			break;
		default:
			break;
		}
	}

	return false;
}

template <typename K, typename V>
void OQHashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize * 3 / 2);
}

template <typename K, typename V>
void OQHashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets
	std::vector<Bucket> bucketsNew(bucketCountNew);

	// Walk through all the current elements and insert them into the new buckets
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
			continue;

		// Hash the key and find the starting bucket
		size_t hash = b->hash;
		size_t iBucketStart = hash % bucketCountNew;

		// Search for an unused bucket
		Bucket * bTarget = nullptr;
		for (size_t j = 0; j < bucketCountNew; ++j) 
		{
			// Evaluate the probing sequence
			size_t probe = (iBucketStart + (j + j*j) / 2) % bucketCountNew;

			Bucket * bNew = &bucketsNew[probe];
			if (bNew->state != BSTATE_Filled)
			{
				bTarget = bNew;
				break;
			}
		}

		assert(bTarget);

		// Store the hash, key, and value in the bucket
		bTarget->hash = hash;
		bTarget->state = BSTATE_Filled;
		bTarget->key = std::move(b->key);
		bTarget->value = std::move(b->value);
	}

	// Swap the new buckets into place
	buckets.swap(bucketsNew);
}

template <typename K, typename V>
void OQHashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);

	size = 0;
}



// DO1HashTable implementation

template <typename K, typename V>
DO1HashTable<K, V>::DO1HashTable()
:	size(0)
{
	// Start off with a small initial size
	buckets.resize(s_hashTableInitialSize);
	keyvals.resize(s_hashTableInitialSize);
}

template <typename K, typename V>
void DO1HashTable<K, V>::Insert(K key, V value)
{
	// Resize larger if the load factor goes over 2/3
	if (size * 3 > buckets.size() * 2)
	{
		Rehash(buckets.size() * 2);
	}

	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search for an unused bucket
	Bucket * bTarget = nullptr;
	KV * kvTarget = nullptr;
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i) 
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
		{
			bTarget = b;
			kvTarget = &keyvals[i];
			break;
		}
	}
	if (!bTarget)
	{
		for (size_t i = 0; i < iBucketStart; ++i)
		{
			Bucket * b = &buckets[i];
			if (b->state != BSTATE_Filled)
			{
				bTarget = b;
				kvTarget = &keyvals[i];
				break;
			}
		}
	}

	assert(bTarget);
	assert(kvTarget);

	// Store the hash, key, and value in the bucket
	bTarget->hash = hash;
	bTarget->state = BSTATE_Filled;
	kvTarget->key = key;
	kvTarget->value = value;

	++size;
}

template <typename K, typename V>
V * DO1HashTable<K, V>::Lookup(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash)
			{
				KV * kv = &keyvals[i];
				if (kv->key == key)
					return &kv->value;
			}
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash)
			{
				KV * kv = &keyvals[i];
				if (kv->key == key)
					return &kv->value;
			}
			break;
		default:
			break;
		}
	}

	return nullptr;
}

template <typename K, typename V>
bool DO1HashTable<K, V>::Remove(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash)
			{
				KV * kv = &keyvals[i];
				if (kv->key == key)
				{
					b->hash = 0;
					b->state = BSTATE_Removed;
					--size;
					return true;
				}
			}
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash)
			{
				KV * kv = &keyvals[i];
				if (kv->key == key)
				{
					b->hash = 0;
					b->state = BSTATE_Removed;
					--size;
					return true;
				}
			}
			break;
		default:
			break;
		}
	}

	return false;
}

template <typename K, typename V>
void DO1HashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize * 3 / 2);
}

template <typename K, typename V>
void DO1HashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets and keyvals
	std::vector<Bucket> bucketsNew(bucketCountNew);
	std::vector<KV> keyvalsNew(bucketCountNew);

	// Walk through all the current elements and insert them into the new buckets
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
			continue;

		// Hash the key and find the starting bucket
		size_t hash = b->hash;
		size_t iBucketStart = hash % bucketCountNew;

		// Search for an unused bucket
		Bucket * bTarget = nullptr;
		KV * kvTarget = nullptr;
		for (size_t j = iBucketStart; j < bucketCountNew; ++j) 
		{
			Bucket * bNew = &bucketsNew[j];
			if (bNew->state != BSTATE_Filled)
			{
				bTarget = bNew;
				kvTarget = &keyvalsNew[j];
				break;
			}
		}
		if (!bTarget)
		{
			for (size_t j = 0; j < iBucketStart; ++j)
			{
				Bucket * bNew = &bucketsNew[j];
				if (bNew->state != BSTATE_Filled)
				{
					bTarget = bNew;
					kvTarget = &keyvalsNew[j];
					break;
				}
			}
		}

		assert(bTarget);
		assert(kvTarget);

		// Store the hash, key, and value in the bucket
		bTarget->hash = hash;
		bTarget->state = BSTATE_Filled;
		KV * kv = &keyvals[i];
		kvTarget->key = std::move(kv->key);
		kvTarget->value = std::move(kv->value);
	}

	// Swap the new buckets and keyvals into place
	buckets.swap(bucketsNew);
	keyvals.swap(keyvalsNew);
}

template <typename K, typename V>
void DO1HashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);
	keyvals.clear();
	keyvals.resize(s_hashTableInitialSize);

	size = 0;
}



// DO2HashTable implementation

template <typename K, typename V>
DO2HashTable<K, V>::DO2HashTable()
:	size(0)
{
	// Start off with a small initial size
	buckets.resize(s_hashTableInitialSize);
	keys.resize(s_hashTableInitialSize);
	values.resize(s_hashTableInitialSize);
}

template <typename K, typename V>
void DO2HashTable<K, V>::Insert(K key, V value)
{
	// Resize larger if the load factor goes over 2/3
	if (size * 3 > buckets.size() * 2)
	{
		Rehash(buckets.size() * 2);
	}

	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search for an unused bucket
	Bucket * bTarget = nullptr;
	size_t iBucketTarget = 0;
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i) 
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
		{
			bTarget = b;
			iBucketTarget = i;
			break;
		}
	}
	if (!bTarget)
	{
		for (size_t i = 0; i < iBucketStart; ++i)
		{
			Bucket * b = &buckets[i];
			if (b->state != BSTATE_Filled)
			{
				bTarget = b;
				iBucketTarget = i;
				break;
			}
		}
	}

	assert(bTarget);

	// Store the hash, key, and value in the bucket
	bTarget->hash = hash;
	bTarget->state = BSTATE_Filled;
	keys[iBucketTarget] = key;
	values[iBucketTarget] = value;

	++size;
}

template <typename K, typename V>
V * DO2HashTable<K, V>::Lookup(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash && keys[i] == key)
				return &values[i];
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return nullptr;
		case BSTATE_Filled:
			if (b->hash == hash && keys[i] == key)
				return &values[i];
			break;
		default:
			break;
		}
	}

	return nullptr;
}

template <typename K, typename V>
bool DO2HashTable<K, V>::Remove(K key)
{
	// Hash the key and find the starting bucket
	size_t hash = HashKey(key) & s_62Bits;
	size_t iBucketStart = hash % buckets.size();

	// Search the buckets until we hit an empty one
	for (size_t i = iBucketStart, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash && keys[i] == key)
			{
				b->hash = 0;
				b->state = BSTATE_Removed;
				--size;
				return true;
			}
			break;
		default:
			break;
		}
	}
	for (size_t i = 0; i < iBucketStart; ++i)
	{
		Bucket * b = &buckets[i];
		switch (b->state)
		{
		case BSTATE_Empty:
			return false;
		case BSTATE_Filled:
			if (b->hash == hash && keys[i] == key)
			{
				b->hash = 0;
				b->state = BSTATE_Removed;
				--size;
				return true;
			}
			break;
		default:
			break;
		}
	}

	return false;
}

template <typename K, typename V>
void DO2HashTable<K, V>::Reserve(size_t maxSize)
{
	Rehash(maxSize * 3 / 2);
}

template <typename K, typename V>
void DO2HashTable<K, V>::Rehash(size_t bucketCountNew)
{
	// Can't rehash down to smaller than current size or initial size
	bucketCountNew = std::max(std::max(bucketCountNew, size),
						   size_t(s_hashTableInitialSize));

	// Build a new set of buckets, keys, and values
	std::vector<Bucket> bucketsNew(bucketCountNew);
	std::vector<K> keysNew(bucketCountNew);
	std::vector<V> valuesNew(bucketCountNew);

	// Walk through all the current elements and insert them into the new buckets
	for (size_t i = 0, iEnd = buckets.size(); i < iEnd; ++i)
	{
		Bucket * b = &buckets[i];
		if (b->state != BSTATE_Filled)
			continue;

		// Hash the key and find the starting bucket
		size_t hash = b->hash;
		size_t iBucketStart = hash % bucketCountNew;

		// Search for an unused bucket
		Bucket * bTarget = nullptr;
		size_t iBucketTarget = 0;
		for (size_t j = iBucketStart; j < bucketCountNew; ++j) 
		{
			Bucket * bNew = &bucketsNew[j];
			if (bNew->state != BSTATE_Filled)
			{
				bTarget = bNew;
				iBucketTarget = j;
				break;
			}
		}
		if (!bTarget)
		{
			for (size_t j = 0; j < iBucketStart; ++j)
			{
				Bucket * bNew = &bucketsNew[j];
				if (bNew->state != BSTATE_Filled)
				{
					bTarget = bNew;
					iBucketTarget = j;
					break;
				}
			}
		}

		assert(bTarget);

		// Store the hash, key, and value in the bucket
		bTarget->hash = hash;
		bTarget->state = BSTATE_Filled;
		keysNew[iBucketTarget] = std::move(keys[i]);
		valuesNew[iBucketTarget] = std::move(values[i]);
	}

	// Swap the new buckets, keys, and values into place
	buckets.swap(bucketsNew);
	keys.swap(keysNew);
	values.swap(valuesNew);
}

template <typename K, typename V>
void DO2HashTable<K, V>::Reset()
{
	// Blow away the current table and reset to small initial size
	buckets.clear();
	buckets.resize(s_hashTableInitialSize);
	keys.clear();
	keys.resize(s_hashTableInitialSize);
	values.clear();
	values.resize(s_hashTableInitialSize);

	size = 0;
}
